# coding: utf-8

"""
    LAMP Platform

    The LAMP Platform API.

    The version of the OpenAPI document: 1.0.0
    Contact: team@digitalpsych.org
"""


from __future__ import absolute_import
import re
import sys

import six
import nulltype

from LAMP.model_utils import (
    ModelComposed,
    ModelNormal,
    ModelSimple,
    date,
    datetime,
    file_type,
    int,
    none_type,
    str,
    validate_get_composed_info,
)


class ActivityEvent(ModelNormal):
    """Attributes:
      allowed_values (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          with a capitalized key describing the allowed value and an allowed
          value. These dicts store the allowed enum values.
      attribute_map (dict): The key is attribute name
          and the value is json key in definition.
      discriminator_value_class_map (dict): A dict to go from the discriminator
          variable value to the discriminator class name.
      validations (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          that stores validations for max_length, min_length, max_items,
          min_items, exclusive_maximum, inclusive_maximum, exclusive_minimum,
          inclusive_minimum, and regex.
      additional_properties_type (tuple): A tuple of classes accepted
          as additional properties values.
    """

    allowed_values = {
    }

    validations = {
    }

    additional_properties_type = None

    @staticmethod
    def openapi_types():
        """
        This must be a class method so a model may have properties that are
        of type self, this ensures that we don't create a cyclic import

        Returns
            openapi_types (dict): The key is attribute name
                and the value is attribute type.
        """
        return {
            'timestamp': (float,),
            'duration': (int,),
            'activity': (str,),
            'data': (bool, date, datetime, dict, float, int, list, str,),
            'temporal_slices': ([object],),
        }

    @staticmethod
    def discriminator():
        return None

    attribute_map = {
        'timestamp': 'timestamp',
        'duration': 'duration',
        'activity': 'activity',
        'data': 'data',
        'temporal_slices': 'temporal_slices',
    }

    @staticmethod
    def _composed_schemas():
        return None

    required_properties = set([
        '_data_store',
        '_check_type',
        '_from_server',
        '_path_to_item',
        '_configuration',
    ])

    def __init__(self, _check_type=True, _from_server=False, _path_to_item=(), _configuration=None, **kwargs):
        """activity_event.ActivityEvent - a model defined in OpenAPI

        Keyword Args:
            _check_type (bool): if True, values for parameters in openapi_types
                                will be type checked and a TypeError will be
                                raised if the wrong type is input.
                                Defaults to True
            _path_to_item (tuple/list): This is a list of keys or values to
                                drill down to the model in received_data
                                when deserializing a response
            _from_server (bool): True if the data is from the server
                                False if the data is from the client (default)
            _configuration (Configuration): the instance to use when
                                deserializing a file_type parameter.
                                If passed, type conversion is attempted
                                If omitted no type conversion is done.
            timestamp (float): The UNIX Epoch date-time representation: number of milliseconds since 1/1/1970 12:00 AM.. [optional]
            duration (int): The duration this event lasted before recording ended.. [optional]
            activity (str): A globally unique reference for objects.. [optional]
            data (bool, date, datetime, dict, float, int, list, str): The summary information for the activity event as determined by the activity that created this activity event.. [optional]
            temporal_slices ([object]): The specific interaction details of the activity event.. [optional]
        """

        self._data_store = {}
        self._check_type = _check_type
        self._from_server = _from_server
        self._path_to_item = _path_to_item
        self._configuration = _configuration

        for var_name, var_value in six.iteritems(kwargs):
            if var_name not in self.attribute_map and \
                        self._configuration is not None and \
                        self._configuration.discard_unknown_keys and \
                        self.additional_properties_type is None:
                # discard variable.
                continue
            setattr(self, var_name, var_value)



import pandas as pd 
import numpy as np
import datetime
import os
import math
import LAMP
import itertools
from functools import reduce


class ParticipantExt():
    """
    Create participant dataframe
    """
    def __init__(self, 
                 id, 
                 domains=None, 
                 age=None, 
                 race=None, 
                 sex=None, 
                 df_props={}):

        self.id = id
        self.domains = domains
        self.age = age
        self.race = race
        self.sex = sex

        self.df = self.create_df(**df_props)

        self.impute_status = False
        self.bin_status = False
        self.normalize_status = False	

    @property
    def id(self):
        return self._id

    @property
    def age(self):
        return self._age

    @property
    def race(self):
        return self._race

    @property
    def sex(self):
        return self._sex

    @id.setter
    def id(self, value):
        self._id = value

    @age.setter
    def age(self, value):
        self._age = value

    @race.setter
    def race(self, value):
        self._race = value

    @sex.setter
    def sex(self, value):
        self._sex = value

    def reset(self):
        """
        Resets the participant's df to be the original version
        """
        self.df = self.create_df()
        self.impute_status, self.bin_status, self.normalize_status = False, False, False

    def domain_check(self, domains):
        """
        If domains is passed in, just return it.
        Else, see if domains is set as object attribute
        """
        if domains is None:
            if not hasattr(self, 'domains'):
                raise AttributeError('Domains were not set for cohort and were not provided.')
            domains = self.domains
        return domains

    def attachment_results(self):
        ATTACHMENTS = {"beta_a", "beta_b", 
                      "screen_features.screen_time", 'screen_features.sessions', 'screen_features.checks', 'screen_features.session_time'}
        
        attachment_dict = {}
        for feature in ATTACHMENTS:
            try:
                attachment_events = LAMP.Type.get_attachment(type_id=self.id, attachment_key="lamp." + feature)["data"]
            except Exception as e: #unable to find resource
                continue
            
            if len(attachment_events) > 0:
                attachment_dict[feature] = [(event['value'], event['timestamp']) for event in attachment_events]
        
        #Sort
        for feature in attachment_dict:
            attachment_dict[feature] = sorted(attachment_dict[feature], key=lambda x: x[1])
        
        return attachment_dict
    
    def passive_feature_results(self, resolution):
        """
        """
        #Find beiwe id
        RESOLUTION_KEY = {'day':'daily', 'week':'weekly', 'month':'monthly'}
        PASSIVE_FEATURES = ['Hometime', 'DistTravelled', 'RoG', 'MaxDiam',
                           'MaxHomeDist', 'SigLocsVisited', 'AvgFlightLen', 'StdFlightLen',
                           'AvgFlightDur', 'StdFlightDur', 'ProbPause', 'SigLocEntropy',
                           'MinsMissing', 'CircdnRtn', 'WkEndDayRtn', 'outgoing_texts',
                           'outgoing_textlengths', 'text_outdegree', 'incoming_texts',
                           'incoming_textlengths', 'text_indegree', 'text_reciprocity',
                           'text_responsiveness', 'outgoing_calls', 'outgoing_calllengths',
                           'call_outdegree', 'incoming_calls', 'incoming_calllengths',
                           'call_indegree', 'call_reciprocity', 'call_responsiveness']

        #Get all passive feature events
        passive_feature_dict = {}
        for feature in PASSIVE_FEATURES:            
            feature_query, feature_query_2 = 'beiwe.' + feature + '.' + RESOLUTION_KEY[resolution], 'beiwe.passive_features.' + feature + '.' + RESOLUTION_KEY[resolution] 

            feature_events, feature_events_2 = LAMP.SensorEvent.all_by_participant(participant_id=self.id, origin=feature_query), LAMP.SensorEvent.all_by_participant(participant_id=self.id, origin=feature_query_2)

            if len(feature_events['data']) > 0:
                passive_feature_dict[feature] = []
                for event in feature_events['data']:
                    passive_feature_dict[feature].append((event['data']['value'], event['timestamp']))
                    
            elif len(feature_events_2['data']) > 0:
                passive_feature_dict[feature] = []
                for event in feature_events_2['data']:
                    passive_feature_dict[feature].append((event['data']['value'], event['timestamp']))

        #Sort
        for feature in passive_feature_dict:
            passive_feature_dict[feature] = sorted(passive_feature_dict[feature], key=lambda x: x[1])
        
        return passive_feature_dict

    def survey_results(self, participant=None, question_categories=None):
        """
        Get survey events for participant
        :param participant (str): the LAMP ID for participant. If not provided, then take participant id
        :param question_categories (bool): indicates whether to use custom question mappings as defined in params file
        """

        if participant is None:
            participant = self.id

        participant_activities = LAMP.Activity.all_by_participant(participant)['data']
        participant_activities_surveys = [activity for activity in participant_activities if activity['spec'] == 'lamp.survey']
        participant_activities_surveys_ids = [survey['id'] for survey in participant_activities_surveys]        

        participant_results = [result for result in LAMP.ActivityEvent.all_by_participant(participant)['data'] if result['activity'] in participant_activities_surveys_ids and len(result['temporal_slices']) > 0]
        participant_surveys = {} #maps survey_type to occurence of scores 
        for result in participant_results:
            #Check if it's a survey event
            if result['activity'] not in participant_activities_surveys_ids or len(result['temporal_slices']) == 0: 
                continue

            activity = LAMP.Activity.view(result['activity'])['data'][0]
            result_settings = activity['settings']

            survey_time = result['timestamp']
            survey_result = {} #maps question domains to scores
            for event in result['temporal_slices']: #individual questions in a survey
                question = event['item']
                
                for i in range(len(result_settings)) : #match question info to question
                    if result_settings[i]['text'] == question: 
                        current_question_info=result_settings[i]
                        break

                #score based on question type:
                score=None
                event_value=event.get('value') #safely get event['value'] to protect from missing keys
                
                if current_question_info['type'] =='likert' and event_value!=None :
                    score = float(event_value)
                        
                elif current_question_info['type']=='boolean':
                    if event_value == 'no': score = 0.0 #no is healthy in standard scoring
                    elif event_value =='yes' : score = 3.0 # yes is healthy in reverse scoring

                elif current_question_info['type'] == 'list' :
                    for option_index in range(len(current_question_info['options'])) :
                        if event_value == current_question_info['options'][option_index] :
                            score = option_index * 3 / (len(current_question_info['options'])-1)

                elif current_question_info['type'] == 'text':  #skip
                    continue
                
                if score==None : continue
                #add event to a category, either user-defined or default activity
                if question_categories:
                    #See if there is an extra space in the string
                    if question not in question_categories:
                        if question[:-1] in question_categories:
                            question = question[:-1]
                        else:
                            continue

                    event_category = question_categories[question]['category']
                    #flip score if necessary
                    if question_categories[question]['reverse_scoring']: 
                        score = 3.0 - score

                    if event_category in survey_result: survey_result[event_category].append(score) 
                    else: survey_result[event_category] = [score]

                else:
                    if activity['name'] not in survey_result:
                        survey_result[activity['name']] = []
                    survey_result[activity['name']].append(score)
                    

            #add mean to each cat to master dictionary           
            for category in survey_result: 
                survey_result[category] = np.mean(survey_result[category])
                if category not in participant_surveys: 
                    participant_surveys[category] = [(survey_result[category], survey_time)]
                else: 
                    participant_surveys[category].append((survey_result[category], survey_time))

        #Sort surveys by time
        for activity_category in participant_surveys:
            participant_surveys[activity_category] = sorted(participant_surveys[activity_category], key=lambda x: x[1])

        return participant_surveys
        

    def create_df(self, days_cap=120, day_first=None, day_last=None, resolution='day', start_monday=False, start_morning=True, time_centered=False, question_categories=None):
        """
        Create participant dataframe
        :param day_first (datetime.Date)
        """

        FIFTEEN_MIN_PER_UNIT = {'15 min': 1, 'day': 4*24, 'week': 4*24*7, 'month': 4*24*30}
        UNITS_PER_DAY = {'15 min': 4*24, 'day': 1, 'week': 1/7, 'month': 1/30}

  
        assert resolution in ['15 min', 'day', 'week', 'month']

        surveys = self.survey_results(question_categories=question_categories) #survey ActivityEvents
        passive_features = self.passive_feature_results(resolution=resolution) #beiewe.passive_features
        attachment_features = self.attachment_results() #static attachment features
        
        surveys = {**surveys, **passive_features, **attachment_features}
        #surveys.update(passive_features).update(attachment_features)

        if len(surveys) == 0:
            return None

        #Find the first, last date
        if day_first is None: day_first = datetime.datetime.utcfromtimestamp(sorted([surveys[dom][0][1]/1000 for dom in surveys])[0])
        else: day_first = datetime.datetime.combine(day_first, datetime.time.min) #convert to datetime

        if day_last is None: day_last = datetime.datetime.utcfromtimestamp(sorted([surveys[dom][-1][1]/1000 for dom in surveys])[-1])
        else: day_last = datetime.datetime.combine(day_last, datetime.time.min) #convert to datetime

        #Clip days based on morning and weekday parameters
        if start_monday:
            if day_first.weekday() > 0: 
                day_first += datetime.timedelta(days = - day_first.weekday())

        if start_morning: 
            day_first, day_last = day_first.replace(hour=9, minute=0, second=0), day_last.replace(hour=9, minute=0, second=0)
        days_elapsed = (day_last - day_first).days 
        date_list = [day_first + datetime.timedelta(minutes=15*FIFTEEN_MIN_PER_UNIT[resolution]*x) for x in range(0, math.ceil(min(days_elapsed, days_cap) * UNITS_PER_DAY[resolution]))]

        #Create dateframe for the number of time units that have data; limited by days; cap at 'days_cap' if this number is large
        df = pd.DataFrame({'Date': date_list, 'id':self.id})

        domains = [dom for dom in surveys]
        for dom in domains: 
            df[dom] = np.nan

        #Parse surveys
        for dom in surveys:
            if dom not in domains and domains is not None:
                continue

            #Based on resolution, match each survey event to its closest date
            dates = [datetime.datetime.utcfromtimestamp(event_time/1000) for _, event_time in surveys[dom] if day_first <= datetime.datetime.utcfromtimestamp(event_time/1000) <= day_last] 

            #Choose closest date if "time centered"; else, choose preceding date
            if time_centered: rounded_dates = [df.loc[df.index[(date - df['Date']).abs().sort_values().index[0]], 'Date'] for date in dates]
            else: rounded_dates = [df.loc[df.index[(date - df['Date'])[(date - df['Date']) >= datetime.timedelta(0)].sort_values().index[0]], 'Date'] for date in dates]

            results = [event_val for event_val, event_time in surveys[dom] if day_first <= datetime.datetime.utcfromtimestamp(event_time/1000) <= day_last]
            dom_results = pd.DataFrame({'Date':dates, 'Rounded Date':rounded_dates, 'Result':results})
            for date, date_df in dom_results.groupby('Rounded Date'):                    
                df.loc[df['Date'] == date.to_pydatetime(), dom] = np.mean(date_df['Result']) 

        #Convert to date to actual date objects if resolution is day or greater
        if resolution != '15 min':
            df['Date'] = df['Date'].apply(lambda d: d.date())
            
        #Trim columns if there are predetermined domains
        if self.domains is not None: 
            df = df.loc[:, ['id', 'Date'] + [d for d in self.domains if d in df.columns.values]]

        return df

 
    def impute(self, domains):
        """
        Get value for each column for each window
        """
        if self.impute_status:
            print('Dataframe already imputed.')
            return

        weighted_dict = [0.05, 0.20, 0.40, 1.5, 0.4, 0.20, 0.05]

        #Get indices of all middle bin values; add them to new df
        for dom in domains:
            if dom not in self.df:
                continue

            dom_values = []

            for ind in range(len(self.df.index)):

                #Get indices
                middle_weight_index = 3
                starting_index = max(ind -3, 0)
                ending_index = min(ind + 4, 90)

                #Get slice values
                subj_slice = self.df.iloc[starting_index:ending_index]

                #Remove na
                subj_slice_no_nan = subj_slice[dom].dropna()
                slice_indices = subj_slice_no_nan.index

                if len(slice_indices) == 0:
                    dom_values.append(np.nan)
                    continue

                #Match slice index with weight index
                weighted_dict_vals = [weighted_dict[middle_weight_index - (ind - slice_i)] for slice_i in slice_indices]

                #Find total in bin
                slice_val = sum(subj_slice_no_nan * [val / sum(weighted_dict_vals) for val in weighted_dict_vals])
                dom_values.append(slice_val)

            self.df[dom] = dom_values

        self.impute_status = True


    def bin(self, domains, window_size=3, shift=0):
        """
        Bin dataframe
        :param domains (list): the domains to bin 
        :window_size (int): the size of the bins (in days)
        :shift (int): the day of the week to start the binning on (Monday == 0)
        """

        #domains = self.domain_check(domains)
        domains = self.df.columns.drop(['Date', 'id'])
        
        #Shift until Monday
        df_copy = self.df.copy()
        if shift is not None:
            try:
                dow = df_copy.iloc[0]['Date'].weekday()
                if dow > 0 and len(df_copy) > dow:
                    df_copy = df_copy.shift(shift - dow)
            except:
                print(self.id, df_copy)
        df_copy['bin'] = np.floor(df_copy.index / window_size )
        bins = df_copy.groupby('bin')
        subj_bin_df = pd.DataFrame(columns=['Bin Start Date', 'Bin End Date'] + domains.values.tolist())
        for b in bins:
            bin_values = []
            #Add bin start/end dates
            
            start_date, end_date = b[1].iloc[0]['Date'], b[1].iloc[-1]['Date']
            bin_values.extend((start_date, end_date))
            for dom in domains:
                if dom in b[1].columns:
                    bin_dom_value = b[1][dom].mean()
                    bin_values.append(bin_dom_value)
                else:
                    bin_values.append(np.nan)

            #Add date range
            subj_bin_df.loc[b[0]] = bin_values    

        subj_bin_df['id'] = self.id
        
        self.bins = subj_bin_df
        
    def impute_bins(self, domains):
        """
        Try to impute bin objects
        """
        assert self.bins is not None
        
        for d in domains:
            for index, row in self.bins.iterrows():
                if 0 < index < len(self.bins.index) - 1:
                    index = int(index)
                    if pd.isnull(self.bins.iloc[index][d]) and not pd.isnull(self.bins.iloc[index-1][d]) and not pd.isnull(self.bins.iloc[index+1][d]):                        
                        self.bins.at[index,d] = np.mean([self.bins.iloc[index-1][d], self.bins.iloc[index+1][d]])


    def normalize(self, domains, domain_means={}, domain_vars={}):
        """
        Normalize columns values to 0 mean/ unit variance
        :param domain_means (dict): the mean for each column value
        :param domain_vars (dict): the variance for each column value
        If mean/var not provided, resort to in-sample normalization
        """
        if self.normalize_status: return
 
        domains = self.domain_check(domains)
        if domain_means == {} and domain_vars == {}:
            for dom in domains:
                if dom in self.df.columns:
                    domain_means[dom] = self.df[dom].mean()
                    domain_vars[dom] = self.df[dom].std()

        for dom in domains:
            if dom in self.df.columns and dom in domain_means and dom in domain_vars:
                self.df[dom] = (self.df[dom] - domain_means[dom]) / domain_vars[dom]

            self.normalize_status = True

    def create_transition_dict(self, level):
        """
        Create nested dictionary structure 
        :param level (int): the level dictionary structure. Must be >= 0
        """
        trans_dict = {}
        for comb in itertools.product(('out', 'in'), repeat=level):
            trans_dict[comb] = {comb2:0 for comb2 in itertools.product(('out', 'in'), repeat=level)}
        return trans_dict


    def assign_transition_dict(self, trans_dict, row, row2):
        """
        Increment transition dict
        """
        label1 = tuple(['in' if col < 1.0 else 'out' for col in row])
        label2 = tuple(['in' if col < 1.0 else 'out' for col in row2])
        trans_dict[label1][label2] += 1

    def get_transitions(self, domains=None, joint_size=1):
        """
        Count transition events for each col in subj_df
        """
        domains = self.domain_check(domains)
        all_trans_dict = {}
        for dom_group in itertools.combinations(domains, r=joint_size):

            #Create trans dictionary
            group_dict = self.create_transition_dict(level=joint_size)

            #Find bins with values for each group
            good_bins = self.bins[list(dom_group)].dropna()

            #Assign
            row_iterator = good_bins.iterrows()
            try:
                last_i, last = next(row_iterator)
            except StopIteration:
                continue
            for index, row in row_iterator:
                if int(index) - int(last_i) <= 3:
                    self.assign_transition_dict(group_dict, last, row)
                last_i, last = index, row

            all_trans_dict[dom_group] = group_dict

        return all_trans_dict

    def domain_bouts(self, domains=None):
        """
        """
        def parse_bout_list(bout_list, state, low_bouts, high_bouts):
            """
            Helper function to parse bout list at end of bout
            """
            if len(bout_list) == 1: bout_list.append(bout_list[-1] + 3) #edge case where last domain event is only one in its bout

            if state: low_bouts.append(float(bout_list[-1]) - float(bout_list[0]))
            else: high_bouts.append(float(bout_list[-1]) - float(bout_list[0]))
            return low_bouts, high_bouts

        domains = self.domain_check(domains)
        bout_dict = {}
        for dom in domains:
            if dom not in self.df:
                continue

            bout_list = [] #temporary list that contains times of current bout
            subj_dom = self.df.loc[self.df[dom].notnull(), dom]
            row_iterator = subj_dom.iteritems()
            try:
                last_day, last_val = next(row_iterator)
                bout_list.append(last_day)
                if last_val < 1.0: last_state = True #set this back on first val
                else: last_state = False
            except StopIteration:
                continue

            bout_dict[dom] = {}
            low_bouts, high_bouts = [], [] #duration of all in-range bouts
            low_bouts_end, high_bouts_end = 0, 0 #counter the keep track of # of ended bout things
            for day, val in row_iterator:
                if val < 1.0: state = True
                else: state = False

                if last_state == state and day - last_day <= 6: #continue bout
                    bout_list.append(day)

                else: #discontinue bout
                    if day - last_day > 8: 
                        bout_list.append(last_day + 3) #If adjacent rows are day outside threshold, discontinue bout;cap last bout at 3 days past last activity	
                        if last_state: low_bouts_end += 1
                        else: high_bouts_end += 1
                    else: bout_list.append(day) #then normal switch

                    low_bouts, high_bouts = parse_bout_list(bout_list, last_state, low_bouts, high_bouts)
                    bout_list = [day]

                last_day, last_val = day, val
                last_state = state

            low_bouts, high_bouts = parse_bout_list(bout_list, last_state, low_bouts, high_bouts) #parse last bout
            bout_dict[dom]['low'], bout_dict[dom]['high'] = [float(b) for b in low_bouts], [float(b) for b in high_bouts]
            bout_dict[dom]['low ends'], bout_dict[dom]['high ends'] = low_bouts_end, high_bouts_end

        return bout_dict


# coding: utf-8

"""
    LAMP Platform

    The LAMP Platform API.

    The version of the OpenAPI document: 1.0.0
    Contact: team@digitalpsych.org
"""


from __future__ import absolute_import

__version__ = "develop"

# import apis into sdk package
from LAMP.api.api_api import APIApi
from LAMP.api.activity_api import ActivityApi
from LAMP.api.activity_event_api import ActivityEventApi
from LAMP.api.activity_spec_api import ActivitySpecApi
from LAMP.api.credential_api import CredentialApi
from LAMP.api.participant_api import ParticipantApi
from LAMP.api.researcher_api import ResearcherApi
from LAMP.api.sensor_api import SensorApi
from LAMP.api.sensor_event_api import SensorEventApi
from LAMP.api.sensor_spec_api import SensorSpecApi
from LAMP.api.study_api import StudyApi
from LAMP.api.type_api import TypeApi

# import ApiClient
from LAMP.api_client import ApiClient

# import Configuration
from LAMP.configuration import Configuration

# import exceptions
from LAMP.exceptions import LAMPException
from LAMP.exceptions import ApiTypeError
from LAMP.exceptions import ApiValueError
from LAMP.exceptions import ApiKeyError
from LAMP.exceptions import ApiException

# import models into sdk package
from LAMP.models.access_citation import AccessCitation
from LAMP.models.activity import Activity
from LAMP.models.activity_event import ActivityEvent
from LAMP.models.activity_spec import ActivitySpec
from LAMP.models.credential import Credential
from LAMP.models.document import Document
from LAMP.models.duration_interval import DurationInterval
from LAMP.models.duration_interval_legacy import DurationIntervalLegacy
from LAMP.models.dynamic_attachment import DynamicAttachment
from LAMP.models.error import Error
from LAMP.models.metadata import Metadata
from LAMP.models.participant import Participant
from LAMP.models.researcher import Researcher
from LAMP.models.sensor import Sensor
from LAMP.models.sensor_event import SensorEvent
from LAMP.models.sensor_spec import SensorSpec
from LAMP.models.study import Study
from LAMP.models.temporal_slice import TemporalSlice

from LAMP.analysis.participant_ext import ParticipantExt
from LAMP.analysis.study_ext import StudyExt

API = APIApi()
Type = TypeApi()
Credential = CredentialApi()
Researcher = ResearcherApi()
Study = StudyApi()
Participant = ParticipantApi()
Activity = ActivityApi()
ActivitySpec = ActivitySpecApi()
ActivityEvent = ActivityEventApi()
Sensor = SensorEventApi()
SensorSpec = SensorEventApi()
SensorEvent = SensorEventApi()

def connect(access_key, secret_key, server_address="api.lamp.digital"):
	client = ApiClient(Configuration(host=f"https://{server_address}", username=access_key, password=secret_key))

	global API
	global Type
	global Credential
	global Researcher
	global Study
	global Participant
	global Activity
	global ActivitySpec
	global ActivityEvent
	global Sensor
	global SensorSpec
	global SensorEvent

	API = APIApi(client)
	Type = TypeApi(client)
	Credential = CredentialApi(client)
	Researcher = ResearcherApi(client)
	Study = StudyApi(client)
	Participant = ParticipantApi(client)
	Activity = ActivityApi(client)
	ActivitySpec = ActivitySpecApi(client)
	ActivityEvent = ActivityEventApi(client)
	Sensor = SensorEventApi(client)
	SensorSpec = SensorEventApi(client)
	SensorEvent = SensorEventApi(client)



# coding: utf-8

"""
    LAMP Platform

    The LAMP Platform API.

    The version of the OpenAPI document: 1.0.0
    Contact: team@digitalpsych.org
"""


from __future__ import absolute_import

import unittest

import LAMP
from LAMP.api.api_api import APIApi
from LAMP.rest import ApiException


class TestAPIApi(unittest.TestCase):
    """APIApi unit test stubs"""

    def setUp(self):
        self.api = LAMP.api.api_api.APIApi()

    def tearDown(self):
        pass

    def test_query(self):
        """Test case for query

        Query the LAMP Database.
        """
        pass

    def test_schema(self):
        """Test case for schema

        View the API schema document.
        """
        pass


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

"""
    LAMP Platform

    The LAMP Platform API.

    The version of the OpenAPI document: 1.0.0
    Contact: team@digitalpsych.org
"""


from __future__ import absolute_import

import unittest

import LAMP
from LAMP.api.type_api import TypeApi
from LAMP.rest import ApiException


class TestTypeApi(unittest.TestCase):
    """TypeApi unit test stubs"""

    def setUp(self):
        self.api = LAMP.api.type_api.TypeApi()

    def tearDown(self):
        pass

    def test_get_attachment(self):
        """Test case for get_attachment

        """
        pass

    def test_get_dynamic_attachment(self):
        """Test case for get_dynamic_attachment

        """
        pass

    def test_list_attachments(self):
        """Test case for list_attachments

        """
        pass

    def test_parent(self):
        """Test case for parent

        Find the owner(s) of the resource.
        """
        pass

    def test_set_attachment(self):
        """Test case for set_attachment

        """
        pass

    def test_set_dynamic_attachment(self):
        """Test case for set_dynamic_attachment

        """
        pass


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

"""
    LAMP Platform

    The LAMP Platform API.

    The version of the OpenAPI document: 1.0.0
    Contact: team@digitalpsych.org
"""


from __future__ import absolute_import

import unittest

import LAMP
from LAMP.api.credential_api import CredentialApi
from LAMP.rest import ApiException


class TestCredentialApi(unittest.TestCase):
    """CredentialApi unit test stubs"""

    def setUp(self):
        self.api = LAMP.api.credential_api.CredentialApi()

    def tearDown(self):
        pass

    def test_create(self):
        """Test case for create

        """
        pass

    def test_delete(self):
        """Test case for delete

        """
        pass

    def test_list(self):
        """Test case for list

        """
        pass

    def test_update(self):
        """Test case for update

        """
        pass


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

"""
    LAMP Platform

    The LAMP Platform API.

    The version of the OpenAPI document: 1.0.0
    Contact: team@digitalpsych.org
"""


from datetime import date, datetime
import inspect
import os
import pprint
import re
import tempfile

from dateutil.parser import parse
import six

from LAMP.exceptions import (
    ApiKeyError,
    ApiTypeError,
    ApiValueError,
)

none_type = type(None)
if six.PY3:
    import io
    file_type = io.IOBase
    # these are needed for when other modules import str and int from here
    str = str
    int = int
else:
    file_type = file
    str_py2 = str
    unicode_py2 = unicode
    long_py2 = long
    int_py2 = int
    # this requires that the future library is installed
    from builtins import int, str


class OpenApiModel(object):
    """The base class for all OpenAPIModels"""

    def set_attribute(self, name, value):
        # this is only used to set properties on self

        path_to_item = []
        if self._path_to_item:
            path_to_item.extend(self._path_to_item)
        path_to_item.append(name)

        openapi_types = self.openapi_types()
        if name in openapi_types:
            required_types_mixed = openapi_types[name]
        elif self.additional_properties_type is None:
            raise ApiKeyError(
                "{0} has no key '{1}'".format(type(self).__name__, name),
                path_to_item
            )
        elif self.additional_properties_type is not None:
            required_types_mixed = self.additional_properties_type

        if get_simple_class(name) != str:
            error_msg = type_error_message(
                var_name=name,
                var_value=name,
                valid_classes=(str,),
                key_type=True
            )
            raise ApiTypeError(
                error_msg,
                path_to_item=path_to_item,
                valid_classes=(str,),
                key_type=True
            )

        if self._check_type:
            value = validate_and_convert_types(
                value, required_types_mixed, path_to_item, self._from_server,
                self._check_type, configuration=self._configuration)
        if (name,) in self.allowed_values:
            check_allowed_values(
                self.allowed_values,
                (name,),
                value
            )
        if (name,) in self.validations:
            check_validations(
                self.validations,
                (name,),
                value
            )
        self.__dict__['_data_store'][name] = value

    def __setitem__(self, name, value):
        """this allows us to set values with instance[field_name] = val"""
        self.__setattr__(name, value)

    def __getitem__(self, name):
        """this allows us to get a value with val = instance[field_name]"""
        return self.__getattr__(name)

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other


class ModelSimple(OpenApiModel):
    """the parent class of models whose type != object in their
    swagger/openapi"""

    def __setattr__(self, name, value):
        """this allows us to set a value with instance.field_name = val"""
        if name in self.required_properties:
            self.__dict__[name] = value
            return

        self.set_attribute(name, value)

    def __getattr__(self, name):
        """this allows us to get a value with val = instance.field_name"""
        if name in self.required_properties:
            return self.__dict__[name]

        if name in self.__dict__['_data_store']:
            return self.__dict__['_data_store'][name]

        path_to_item = []
        if self._path_to_item:
            path_to_item.extend(self._path_to_item)
        path_to_item.append(name)
        raise ApiKeyError(
            "{0} has no key '{1}'".format(type(self).__name__, name),
            [name]
        )

    def to_str(self):
        """Returns the string representation of the model"""
        return str(self.value)

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, self.__class__):
            return False

        this_val = self._data_store['value']
        that_val = other._data_store['value']
        types = set()
        types.add(this_val.__class__)
        types.add(that_val.__class__)
        vals_equal = this_val == that_val
        if not six.PY3 and len(types) == 2 and unicode in types:
            vals_equal = (
                this_val.encode('utf-8') == that_val.encode('utf-8')
            )
        if not vals_equal:
            return False
        return True


class ModelNormal(OpenApiModel):
    """the parent class of models whose type == object in their
    swagger/openapi"""

    def __setattr__(self, name, value):
        """this allows us to set a value with instance.field_name = val"""
        if name in self.required_properties:
            self.__dict__[name] = value
            return

        self.set_attribute(name, value)

    def __getattr__(self, name):
        """this allows us to get a value with val = instance.field_name"""
        if name in self.required_properties:
            return self.__dict__[name]

        if name in self.__dict__['_data_store']:
            return self.__dict__['_data_store'][name]

        path_to_item = []
        if self._path_to_item:
            path_to_item.extend(self._path_to_item)
        path_to_item.append(name)
        raise ApiKeyError(
            "{0} has no key '{1}'".format(type(self).__name__, name),
            [name]
        )

    def to_dict(self):
        """Returns the model properties as a dict"""
        return model_to_dict(self, serialize=False)

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, self.__class__):
            return False

        if not set(self._data_store.keys()) == set(other._data_store.keys()):
            return False
        for _var_name, this_val in six.iteritems(self._data_store):
            that_val = other._data_store[_var_name]
            types = set()
            types.add(this_val.__class__)
            types.add(that_val.__class__)
            vals_equal = this_val == that_val
            if (not six.PY3 and
                    len(types) == 2 and unicode in types):
                vals_equal = (
                    this_val.encode('utf-8') == that_val.encode('utf-8')
                )
            if not vals_equal:
                return False
        return True


class ModelComposed(OpenApiModel):
    """the parent class of models whose type == object in their
    swagger/openapi and have oneOf/allOf/anyOf"""

    def __setattr__(self, name, value):
        """this allows us to set a value with instance.field_name = val"""
        if name in self.required_properties:
            self.__dict__[name] = value
            return

        # set the attribute on the correct instance
        model_instances = self._var_name_to_model_instances.get(
            name, self._additional_properties_model_instances)
        if model_instances:
            for model_instance in model_instances:
                if model_instance == self:
                    self.set_attribute(name, value)
                else:
                    setattr(model_instance, name, value)
                if name not in self._var_name_to_model_instances:
                    # we assigned an additional property
                    self.__dict__['_var_name_to_model_instances'][name] = (
                        model_instance
                    )
            return None

        path_to_item = []
        if self._path_to_item:
            path_to_item.extend(self._path_to_item)
        path_to_item.append(name)
        raise ApiKeyError(
            "{0} has no key '{1}'".format(type(self).__name__, name),
            path_to_item
        )

    def __getattr__(self, name):
        """this allows us to get a value with val = instance.field_name"""
        if name in self.required_properties:
            return self.__dict__[name]

        # get the attribute from the correct instance
        model_instances = self._var_name_to_model_instances.get(
            name, self._additional_properties_model_instances)
        path_to_item = []
        if self._path_to_item:
            path_to_item.extend(self._path_to_item)
        path_to_item.append(name)
        values = set()
        if model_instances:
            for model_instance in model_instances:
                if name in model_instance._data_store:
                    values.add(model_instance._data_store[name])
        len_values = len(values)
        if len_values == 0:
            raise ApiKeyError(
                "{0} has no key '{1}'".format(type(self).__name__, name),
                path_to_item
            )
        elif len_values == 1:
            return list(values)[0]
        elif len_values > 1:
            raise ApiValueError(
                "Values stored for property {0} in {1} difffer when looking "
                "at self and self's composed instances. All values must be "
                "the same".format(name, type(self).__name__),
                path_to_item
            )

    def to_dict(self):
        """Returns the model properties as a dict"""
        return model_to_dict(self, serialize=False)

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, self.__class__):
            return False

        if not set(self._data_store.keys()) == set(other._data_store.keys()):
            return False
        for _var_name, this_val in six.iteritems(self._data_store):
            that_val = other._data_store[_var_name]
            types = set()
            types.add(this_val.__class__)
            types.add(that_val.__class__)
            vals_equal = this_val == that_val
            if (not six.PY3 and
                    len(types) == 2 and unicode in types):
                vals_equal = (
                    this_val.encode('utf-8') == that_val.encode('utf-8')
                )
            if not vals_equal:
                return False
        return True


COERCION_INDEX_BY_TYPE = {
    ModelComposed: 0,
    ModelNormal: 1,
    ModelSimple: 2,
    none_type: 3,
    list: 4,
    dict: 5,
    float: 6,
    int: 7,
    bool: 8,
    datetime: 9,
    date: 10,
    str: 11,
    file_type: 12,
}

# these are used to limit what type conversions we try to do
# when we have a valid type already and we want to try converting
# to another type
UPCONVERSION_TYPE_PAIRS = (
    (str, datetime),
    (str, date),
    (list, ModelComposed),
    (dict, ModelComposed),
    (list, ModelNormal),
    (dict, ModelNormal),
    (str, ModelSimple),
    (int, ModelSimple),
    (float, ModelSimple),
    (list, ModelSimple),
)

COERCIBLE_TYPE_PAIRS = {
    False: (  # client instantiation of a model with client data
        # (dict, ModelComposed),
        # (list, ModelComposed),
        # (dict, ModelNormal),
        # (list, ModelNormal),
        # (str, ModelSimple),
        # (int, ModelSimple),
        # (float, ModelSimple),
        # (list, ModelSimple),
        # (str, int),
        # (str, float),
        # (str, datetime),
        # (str, date),
        # (int, str),
        # (float, str),
    ),
    True: (  # server -> client data
        (dict, ModelComposed),
        (list, ModelComposed),
        (dict, ModelNormal),
        (list, ModelNormal),
        (str, ModelSimple),
        (int, ModelSimple),
        (float, ModelSimple),
        (list, ModelSimple),
        # (str, int),
        # (str, float),
        (str, datetime),
        (str, date),
        # (int, str),
        # (float, str),
        (str, file_type)
    ),
}


def get_simple_class(input_value):
    """Returns an input_value's simple class that we will use for type checking
    Python2:
    float and int will return int, where int is the python3 int backport
    str and unicode will return str, where str is the python3 str backport
    Note: float and int ARE both instances of int backport
    Note: str_py2 and unicode_py2 are NOT both instances of str backport

    Args:
        input_value (class/class_instance): the item for which we will return
                                            the simple class
    """
    if isinstance(input_value, type):
        # input_value is a class
        return input_value
    elif isinstance(input_value, tuple):
        return tuple
    elif isinstance(input_value, list):
        return list
    elif isinstance(input_value, dict):
        return dict
    elif isinstance(input_value, none_type):
        return none_type
    elif isinstance(input_value, file_type):
        return file_type
    elif isinstance(input_value, bool):
        # this must be higher than the int check because
        # isinstance(True, int) == True
        return bool
    elif isinstance(input_value, int):
        # for python2 input_value==long_instance -> return int
        # where int is the python3 int backport
        return int
    elif isinstance(input_value, datetime):
        # this must be higher than the date check because
        # isinstance(datetime_instance, date) == True
        return datetime
    elif isinstance(input_value, date):
        return date
    elif (six.PY2 and isinstance(input_value, (str_py2, unicode_py2, str)) or
            isinstance(input_value, str)):
        return str
    return type(input_value)


def check_allowed_values(allowed_values, input_variable_path, input_values):
    """Raises an exception if the input_values are not allowed

    Args:
        allowed_values (dict): the allowed_values dict
        input_variable_path (tuple): the path to the input variable
        input_values (list/str/int/float/date/datetime): the values that we
            are checking to see if they are in allowed_values
    """
    these_allowed_values = list(allowed_values[input_variable_path].values())
    if (isinstance(input_values, list)
            and not set(input_values).issubset(
                set(these_allowed_values))):
        invalid_values = ", ".join(
            map(str, set(input_values) - set(these_allowed_values))),
        raise ApiValueError(
            "Invalid values for `%s` [%s], must be a subset of [%s]" %
            (
                input_variable_path[0],
                invalid_values,
                ", ".join(map(str, these_allowed_values))
            )
        )
    elif (isinstance(input_values, dict)
            and not set(
                input_values.keys()).issubset(set(these_allowed_values))):
        invalid_values = ", ".join(
            map(str, set(input_values.keys()) - set(these_allowed_values)))
        raise ApiValueError(
            "Invalid keys in `%s` [%s], must be a subset of [%s]" %
            (
                input_variable_path[0],
                invalid_values,
                ", ".join(map(str, these_allowed_values))
            )
        )
    elif (not isinstance(input_values, (list, dict))
            and input_values not in these_allowed_values):
        raise ApiValueError(
            "Invalid value for `%s` (%s), must be one of %s" %
            (
                input_variable_path[0],
                input_values,
                these_allowed_values
            )
        )


def check_validations(validations, input_variable_path, input_values):
    """Raises an exception if the input_values are invalid

    Args:
        validations (dict): the validation dictionary
        input_variable_path (tuple): the path to the input variable
        input_values (list/str/int/float/date/datetime): the values that we
            are checking
    """
    current_validations = validations[input_variable_path]
    if ('max_length' in current_validations and
            len(input_values) > current_validations['max_length']):
        raise ApiValueError(
            "Invalid value for `%s`, length must be less than or equal to "
            "`%s`" % (
                input_variable_path[0],
                current_validations['max_length']
            )
        )

    if ('min_length' in current_validations and
            len(input_values) < current_validations['min_length']):
        raise ApiValueError(
            "Invalid value for `%s`, length must be greater than or equal to "
            "`%s`" % (
                input_variable_path[0],
                current_validations['min_length']
            )
        )

    if ('max_items' in current_validations and
            len(input_values) > current_validations['max_items']):
        raise ApiValueError(
            "Invalid value for `%s`, number of items must be less than or "
            "equal to `%s`" % (
                input_variable_path[0],
                current_validations['max_items']
            )
        )

    if ('min_items' in current_validations and
            len(input_values) < current_validations['min_items']):
        raise ValueError(
            "Invalid value for `%s`, number of items must be greater than or "
            "equal to `%s`" % (
                input_variable_path[0],
                current_validations['min_items']
            )
        )

    items = ('exclusive_maximum', 'inclusive_maximum', 'exclusive_minimum',
             'inclusive_minimum')
    if (any(item in current_validations for item in items)):
        if isinstance(input_values, list):
            max_val = max(input_values)
            min_val = min(input_values)
        elif isinstance(input_values, dict):
            max_val = max(input_values.values())
            min_val = min(input_values.values())
        else:
            max_val = input_values
            min_val = input_values

    if ('exclusive_maximum' in current_validations and
            max_val >= current_validations['exclusive_maximum']):
        raise ApiValueError(
            "Invalid value for `%s`, must be a value less than `%s`" % (
                input_variable_path[0],
                current_validations['exclusive_maximum']
            )
        )

    if ('inclusive_maximum' in current_validations and
            max_val > current_validations['inclusive_maximum']):
        raise ApiValueError(
            "Invalid value for `%s`, must be a value less than or equal to "
            "`%s`" % (
                input_variable_path[0],
                current_validations['inclusive_maximum']
            )
        )

    if ('exclusive_minimum' in current_validations and
            min_val <= current_validations['exclusive_minimum']):
        raise ApiValueError(
            "Invalid value for `%s`, must be a value greater than `%s`" %
            (
                input_variable_path[0],
                current_validations['exclusive_maximum']
            )
        )

    if ('inclusive_minimum' in current_validations and
            min_val < current_validations['inclusive_minimum']):
        raise ApiValueError(
            "Invalid value for `%s`, must be a value greater than or equal "
            "to `%s`" % (
                input_variable_path[0],
                current_validations['inclusive_minimum']
            )
        )
    flags = current_validations.get('regex', {}).get('flags', 0)
    if ('regex' in current_validations and
            not re.search(current_validations['regex']['pattern'],
                          input_values, flags=flags)):
        raise ApiValueError(
            r"Invalid value for `%s`, must be a follow pattern or equal to "
            r"`%s` with flags=`%s`" % (
                input_variable_path[0],
                current_validations['regex']['pattern'],
                flags
            )
          )


def order_response_types(required_types):
    """Returns the required types sorted in coercion order

    Args:
        required_types (list/tuple): collection of classes or instance of
            list or dict with classs information inside it

    Returns:
        (list): coercion order sorted collection of classes or instance
            of list or dict with classs information inside it
    """

    def index_getter(class_or_instance):
        if isinstance(class_or_instance, list):
            return COERCION_INDEX_BY_TYPE[list]
        elif isinstance(class_or_instance, dict):
            return COERCION_INDEX_BY_TYPE[dict]
        elif (inspect.isclass(class_or_instance)
                and issubclass(class_or_instance, ModelComposed)):
            return COERCION_INDEX_BY_TYPE[ModelComposed]
        elif (inspect.isclass(class_or_instance)
                and issubclass(class_or_instance, ModelNormal)):
            return COERCION_INDEX_BY_TYPE[ModelNormal]
        elif (inspect.isclass(class_or_instance)
                and issubclass(class_or_instance, ModelSimple)):
            return COERCION_INDEX_BY_TYPE[ModelSimple]
        return COERCION_INDEX_BY_TYPE[class_or_instance]

    sorted_types = sorted(
        required_types,
        key=lambda class_or_instance: index_getter(class_or_instance)
    )
    return sorted_types


def remove_uncoercible(required_types_classes, current_item, from_server,
                       must_convert=True):
    """Only keeps the type conversions that are possible

    Args:
        required_types_classes (tuple): tuple of classes that are required
                          these should be ordered by COERCION_INDEX_BY_TYPE
        from_server (bool): a boolean of whether the data is from the server
                          if false, the data is from the client
        current_item (any): the current item to be converted

    Keyword Args:
        must_convert (bool): if True the item to convert is of the wrong
                          type and we want a big list of coercibles
                          if False, we want a limited list of coercibles

    Returns:
        (list): the remaining coercible required types, classes only
    """
    current_type_simple = get_simple_class(current_item)

    results_classes = []
    for required_type_class in required_types_classes:
        # convert our models to OpenApiModel
        required_type_class_simplified = required_type_class
        if isinstance(required_type_class_simplified, type):
            if issubclass(required_type_class_simplified, ModelComposed):
                required_type_class_simplified = ModelComposed
            elif issubclass(required_type_class_simplified, ModelNormal):
                required_type_class_simplified = ModelNormal
            elif issubclass(required_type_class_simplified, ModelSimple):
                required_type_class_simplified = ModelSimple

        if required_type_class_simplified == current_type_simple:
            # don't consider converting to one's own class
            continue

        class_pair = (current_type_simple, required_type_class_simplified)
        if must_convert and class_pair in COERCIBLE_TYPE_PAIRS[from_server]:
            results_classes.append(required_type_class)
        elif class_pair in UPCONVERSION_TYPE_PAIRS:
            results_classes.append(required_type_class)
    return results_classes


def get_required_type_classes(required_types_mixed):
    """Converts the tuple required_types into a tuple and a dict described
    below

    Args:
        required_types_mixed (tuple/list): will contain either classes or
            instance of list or dict

    Returns:
        (valid_classes, dict_valid_class_to_child_types_mixed):
            valid_classes (tuple): the valid classes that the current item
                                   should be
            dict_valid_class_to_child_types_mixed (doct):
                valid_class (class): this is the key
                child_types_mixed (list/dict/tuple): describes the valid child
                    types
    """
    valid_classes = []
    child_req_types_by_current_type = {}
    for required_type in required_types_mixed:
        if isinstance(required_type, list):
            valid_classes.append(list)
            child_req_types_by_current_type[list] = required_type
        elif isinstance(required_type, tuple):
            valid_classes.append(tuple)
            child_req_types_by_current_type[tuple] = required_type
        elif isinstance(required_type, dict):
            valid_classes.append(dict)
            child_req_types_by_current_type[dict] = required_type[str]
        else:
            valid_classes.append(required_type)
    return tuple(valid_classes), child_req_types_by_current_type


def change_keys_js_to_python(input_dict, model_class):
    """
    Converts from javascript_key keys in the input_dict to python_keys in
    the output dict using the mapping in model_class
    """

    output_dict = {}
    reversed_attr_map = {value: key for key, value in
                         six.iteritems(model_class.attribute_map)}
    for javascript_key, value in six.iteritems(input_dict):
        python_key = reversed_attr_map.get(javascript_key)
        if python_key is None:
            # if the key is unknown, it is in error or it is an
            # additionalProperties variable
            python_key = javascript_key
        output_dict[python_key] = value
    return output_dict


def get_type_error(var_value, path_to_item, valid_classes, key_type=False):
    error_msg = type_error_message(
        var_name=path_to_item[-1],
        var_value=var_value,
        valid_classes=valid_classes,
        key_type=key_type
    )
    return ApiTypeError(
        error_msg,
        path_to_item=path_to_item,
        valid_classes=valid_classes,
        key_type=key_type
    )


def deserialize_primitive(data, klass, path_to_item):
    """Deserializes string to primitive type.

    :param data: str/int/float
    :param klass: str/class the class to convert to

    :return: int, float, str, bool, date, datetime
    """
    additional_message = ""
    try:
        if klass in {datetime, date}:
            additional_message = (
                "If you need your parameter to have a fallback "
                "string value, please set its type as `type: {}` in your "
                "spec. That allows the value to be any type. "
            )
            if klass == datetime:
                if len(data) < 8:
                    raise ValueError("This is not a datetime")
                # The string should be in iso8601 datetime format.
                parsed_datetime = parse(data)
                date_only = (
                    parsed_datetime.hour == 0 and
                    parsed_datetime.minute == 0 and
                    parsed_datetime.second == 0 and
                    parsed_datetime.tzinfo is None and
                    8 <= len(data) <= 10
                )
                if date_only:
                    raise ValueError("This is a date, not a datetime")
                return parsed_datetime
            elif klass == date:
                if len(data) < 8:
                    raise ValueError("This is not a date")
                return parse(data).date()
        else:
            converted_value = klass(data)
            if isinstance(data, str) and klass == float:
                if str(converted_value) != data:
                    # '7' -> 7.0 -> '7.0' != '7'
                    raise ValueError('This is not a float')
            return converted_value
    except (OverflowError, ValueError):
        # parse can raise OverflowError
        raise ApiValueError(
            "{0}Failed to parse {1} as {2}".format(
                additional_message, repr(data), get_py3_class_name(klass)
            ),
            path_to_item=path_to_item
        )


def fix_model_input_data(model_data, model_class):
    # this is only called on classes where the input data is a dict
    fixed_model_data = change_keys_js_to_python(
        model_data,
        model_class
    )
    if model_class._composed_schemas() is not None:
        for allof_class in model_class._composed_schemas()['allOf']:
            fixed_model_data = change_keys_js_to_python(
                fixed_model_data,
                allof_class
            )
    return fixed_model_data


def deserialize_model(model_data, model_class, path_to_item, check_type,
                      configuration, from_server):
    """Deserializes model_data to model instance.

    Args:
        model_data (list/dict): data to instantiate the model
        model_class (OpenApiModel): the model class
        path_to_item (list): path to the model in the received data
        check_type (bool): whether to check the data tupe for the values in
            the model
        configuration (Configuration): the instance to use to convert files
        from_server (bool): True if the data is from the server
            False if the data is from the client

    Returns:
        model instance

    Raise:
        ApiTypeError
        ApiValueError
        ApiKeyError
    """

    kw_args = dict(_check_type=check_type,
                   _path_to_item=path_to_item,
                   _configuration=configuration,
                   _from_server=from_server)

    used_model_class = model_class
    if model_class.discriminator() is not None:
        used_model_class = model_class.get_discriminator_class(
            from_server, model_data)

    if issubclass(used_model_class, ModelSimple):
        instance = used_model_class(value=model_data, **kw_args)
        return instance
    if isinstance(model_data, list):
        instance = used_model_class(*model_data, **kw_args)
    if isinstance(model_data, dict):
        fixed_model_data = change_keys_js_to_python(
            model_data,
            used_model_class
        )
        kw_args.update(fixed_model_data)
        instance = used_model_class(**kw_args)
    return instance


def deserialize_file(response_data, configuration, content_disposition=None):
    """Deserializes body to file

    Saves response body into a file in a temporary folder,
    using the filename from the `Content-Disposition` header if provided.

    Args:
        param response_data (str):  the file data to write
        configuration (Configuration): the instance to use to convert files

    Keyword Args:
        content_disposition (str):  the value of the Content-Disposition
            header

    Returns:
        (file_type): the deserialized file which is open
            The user is responsible for closing and reading the file
    """
    fd, path = tempfile.mkstemp(dir=configuration.temp_folder_path)
    os.close(fd)
    os.remove(path)

    if content_disposition:
        filename = re.search(r'filename=[\'"]?([^\'"\s]+)[\'"]?',
                             content_disposition).group(1)
        path = os.path.join(os.path.dirname(path), filename)

    with open(path, "wb") as f:
        if six.PY3 and isinstance(response_data, str):
            # in python3 change str to bytes so we can write it
            response_data = response_data.encode('utf-8')
        f.write(response_data)

    f = open(path, "rb")
    return f


def attempt_convert_item(input_value, valid_classes, path_to_item,
                         configuration, from_server, key_type=False,
                         must_convert=False, check_type=True):
    """
    Args:
        input_value (any): the data to convert
        valid_classes (any): the classes that are valid
        path_to_item (list): the path to the item to convert
        configuration (Configuration): the instance to use to convert files
        from_server (bool): True if data is from the server, False is data is
            from the client
        key_type (bool): if True we need to convert a key type (not supported)
        must_convert (bool): if True we must convert
        check_type (bool): if True we check the type or the returned data in
            ModelComposed/ModelNormal/ModelSimple instances

    Returns:
        instance (any) the fixed item

    Raises:
        ApiTypeError
        ApiValueError
        ApiKeyError
    """
    valid_classes_ordered = order_response_types(valid_classes)
    valid_classes_coercible = remove_uncoercible(
        valid_classes_ordered, input_value, from_server)
    if not valid_classes_coercible or key_type:
        # we do not handle keytype errors, json will take care
        # of this for us
        if configuration is None or not configuration.discard_unknown_keys:
            raise get_type_error(input_value, path_to_item, valid_classes,
                                 key_type=key_type)
    for valid_class in valid_classes_coercible:
        try:
            if issubclass(valid_class, OpenApiModel):
                return deserialize_model(input_value, valid_class,
                                         path_to_item, check_type,
                                         configuration, from_server)
            elif valid_class == file_type:
                return deserialize_file(input_value, configuration)
            return deserialize_primitive(input_value, valid_class,
                                         path_to_item)
        except (ApiTypeError, ApiValueError, ApiKeyError) as conversion_exc:
            if must_convert:
                raise conversion_exc
            # if we have conversion errors when must_convert == False
            # we ignore the exception and move on to the next class
            continue
    # we were unable to convert, must_convert == False
    return input_value


def validate_and_convert_types(input_value, required_types_mixed, path_to_item,
                               from_server, _check_type, configuration=None):
    """Raises a TypeError is there is a problem, otherwise returns value

    Args:
        input_value (any): the data to validate/convert
        required_types_mixed (list/dict/tuple): A list of
            valid classes, or a list tuples of valid classes, or a dict where
            the value is a tuple of value classes
        path_to_item: (list) the path to the data being validated
            this stores a list of keys or indices to get to the data being
            validated
        from_server (bool): True if data is from the server
            False if data is from the client
        _check_type: (boolean) if true, type will be checked and conversion
            will be attempted.
        configuration: (Configuration): the configuration class to use
            when converting file_type items.
            If passed, conversion will be attempted when possible
            If not passed, no conversions will be attempted and
            exceptions will be raised

    Returns:
        the correctly typed value

    Raises:
        ApiTypeError
    """
    results = get_required_type_classes(required_types_mixed)
    valid_classes, child_req_types_by_current_type = results

    input_class_simple = get_simple_class(input_value)
    valid_type = input_class_simple in set(valid_classes)
    if not valid_type:
        if configuration:
            # if input_value is not valid_type try to convert it
            converted_instance = attempt_convert_item(
                input_value,
                valid_classes,
                path_to_item,
                configuration,
                from_server,
                key_type=False,
                must_convert=True
            )
            return converted_instance
        else:
            raise get_type_error(input_value, path_to_item, valid_classes,
                                 key_type=False)

    # input_value's type is in valid_classes
    if len(valid_classes) > 1 and configuration:
        # there are valid classes which are not the current class
        valid_classes_coercible = remove_uncoercible(
            valid_classes, input_value, from_server, must_convert=False)
        if valid_classes_coercible:
            converted_instance = attempt_convert_item(
                input_value,
                valid_classes_coercible,
                path_to_item,
                configuration,
                from_server,
                key_type=False,
                must_convert=False
            )
            return converted_instance

    if child_req_types_by_current_type == {}:
        # all types are of the required types and there are no more inner
        # variables left to look at
        return input_value
    inner_required_types = child_req_types_by_current_type.get(
        type(input_value)
    )
    if inner_required_types is None:
        # for this type, there are not more inner variables left to look at
        return input_value
    if isinstance(input_value, list):
        if input_value == []:
            # allow an empty list
            return input_value
        for index, inner_value in enumerate(input_value):
            inner_path = list(path_to_item)
            inner_path.append(index)
            input_value[index] = validate_and_convert_types(
                inner_value,
                inner_required_types,
                inner_path,
                from_server,
                _check_type,
                configuration=configuration
            )
    elif isinstance(input_value, dict):
        if input_value == {}:
            # allow an empty dict
            return input_value
        for inner_key, inner_val in six.iteritems(input_value):
            inner_path = list(path_to_item)
            inner_path.append(inner_key)
            if get_simple_class(inner_key) != str:
                raise get_type_error(inner_key, inner_path, valid_classes,
                                     key_type=True)
            input_value[inner_key] = validate_and_convert_types(
                inner_val,
                inner_required_types,
                inner_path,
                from_server,
                _check_type,
                configuration=configuration
            )
    return input_value


def model_to_dict(model_instance, serialize=True):
    """Returns the model properties as a dict

    Args:
        model_instance (one of your model instances): the model instance that
            will be converted to a dict.

    Keyword Args:
        serialize (bool): if True, the keys in the dict will be values from
            attribute_map
    """
    result = {}

    model_instances = [model_instance]
    if model_instance._composed_schemas() is not None:
        model_instances.extend(model_instance._composed_instances)
    for model_instance in model_instances:
        for attr, value in six.iteritems(model_instance._data_store):
            if serialize:
                # we use get here because additional property key names do not
                # exist in attribute_map
                attr = model_instance.attribute_map.get(attr, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: model_to_dict(x, serialize=serialize)
                    if hasattr(x, '_data_store') else x, value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0],
                                  model_to_dict(item[1], serialize=serialize))
                    if hasattr(item[1], '_data_store') else item,
                    value.items()
                ))
            elif hasattr(value, '_data_store'):
                result[attr] = model_to_dict(value, serialize=serialize)
            else:
                result[attr] = value

    return result


def type_error_message(var_value=None, var_name=None, valid_classes=None,
                       key_type=None):
    """
    Keyword Args:
        var_value (any): the variable which has the type_error
        var_name (str): the name of the variable which has the typ error
        valid_classes (tuple): the accepted classes for current_item's
                                  value
        key_type (bool): False if our value is a value in a dict
                         True if it is a key in a dict
                         False if our item is an item in a list
    """
    key_or_value = 'value'
    if key_type:
        key_or_value = 'key'
    valid_classes_phrase = get_valid_classes_phrase(valid_classes)
    msg = (
        "Invalid type for variable '{0}'. Required {1} type {2} and "
        "passed type was {3}".format(
            var_name,
            key_or_value,
            valid_classes_phrase,
            type(var_value).__name__,
        )
    )
    return msg


def get_valid_classes_phrase(input_classes):
    """Returns a string phrase describing what types are allowed
    Note: Adds the extra valid classes in python2
    """
    all_classes = list(input_classes)
    if six.PY2 and str in input_classes:
        all_classes.extend([str_py2, unicode_py2])
    if six.PY2 and int in input_classes:
        all_classes.extend([int_py2, long_py2])
    all_classes = sorted(all_classes, key=lambda cls: cls.__name__)
    all_class_names = [cls.__name__ for cls in all_classes]
    if len(all_class_names) == 1:
        return 'is {0}'.format(all_class_names[0])
    return "is one of [{0}]".format(", ".join(all_class_names))


def get_py3_class_name(input_class):
    if six.PY2:
        if input_class == str:
            return 'str'
        elif input_class == int:
            return 'int'
    return input_class.__name__


def get_allof_instances(self, model_args, constant_args):
    """
    Args:
        self: the class we are handling
        model_args (dict): var_name to var_value
            used to make instances
        constant_args (dict): var_name to var_value
            used to make instances

    Returns
        composed_instances (list)
    """
    composed_instances = []
    for allof_class in self._composed_schemas()['allOf']:

        # transform js keys to python keys in fixed_model_args
        fixed_model_args = change_keys_js_to_python(
            model_args, allof_class)

        # extract a dict of only required keys from fixed_model_args
        kwargs = {}
        var_names = set(allof_class.openapi_types().keys())
        for var_name in var_names:
            if var_name in fixed_model_args:
                kwargs[var_name] = fixed_model_args[var_name]

        # and use it to make the instance
        kwargs.update(constant_args)
        allof_instance = allof_class(**kwargs)
        composed_instances.append(allof_instance)
    return composed_instances


def get_oneof_instance(self, model_args, constant_args):
    """
    Args:
        self: the class we are handling
        model_args (dict): var_name to var_value
            used to make instances
        constant_args (dict): var_name to var_value
            used to make instances

    Returns
        oneof_instance (instance/None)
    """
    if len(self._composed_schemas()['oneOf']) == 0:
        return None

    oneof_instances = []
    for oneof_class in self._composed_schemas()['oneOf']:
        # transform js keys to python keys in fixed_model_args
        fixed_model_args = change_keys_js_to_python(
            model_args, oneof_class)

        # extract a dict of only required keys from fixed_model_args
        kwargs = {}
        var_names = set(oneof_class.openapi_types().keys())
        for var_name in var_names:
            if var_name in fixed_model_args:
                kwargs[var_name] = fixed_model_args[var_name]

        # do not try to make a model with no input args
        if len(kwargs) == 0:
            continue

        # and use it to make the instance
        kwargs.update(constant_args)
        try:
            oneof_instance = oneof_class(**kwargs)
            oneof_instances.append(oneof_instance)
        except Exception:
            pass
    if len(oneof_instances) == 0:
        raise ApiValueError(
            "Invalid inputs given to generate an instance of %s. Unable to "
            "make any instances of the classes in oneOf definition." %
            self.__class__.__name__
        )
    elif len(oneof_instances) > 1:
        raise ApiValueError(
            "Invalid inputs given to generate an instance of %s. Multiple "
            "oneOf instances were generated when a max of one is allowed." %
            self.__class__.__name__
        )
    return oneof_instances[0]


def get_anyof_instances(self, model_args, constant_args):
    """
    Args:
        self: the class we are handling
        model_args (dict): var_name to var_value
            used to make instances
        constant_args (dict): var_name to var_value
            used to make instances

    Returns
        anyof_instances (list)
    """
    anyof_instances = []
    if len(self._composed_schemas()['anyOf']) == 0:
        return anyof_instances

    for anyof_class in self._composed_schemas()['anyOf']:
        # transform js keys to python keys in fixed_model_args
        fixed_model_args = change_keys_js_to_python(model_args, anyof_class)

        # extract a dict of only required keys from these_model_vars
        kwargs = {}
        var_names = set(anyof_class.openapi_types().keys())
        for var_name in var_names:
            if var_name in fixed_model_args:
                kwargs[var_name] = fixed_model_args[var_name]

        # do not try to make a model with no input args
        if len(kwargs) == 0:
            continue

        # and use it to make the instance
        kwargs.update(constant_args)
        try:
            anyof_instance = anyof_class(**kwargs)
            anyof_instances.append(anyof_instance)
        except Exception:
            pass
    if len(anyof_instances) == 0:
        raise ApiValueError(
            "Invalid inputs given to generate an instance of %s. Unable to "
            "make any instances of the classes in anyOf definition." %
            self.__class__.__name__
        )
    return anyof_instances


def get_additional_properties_model_instances(
        composed_instances, self):
    additional_properties_model_instances = []
    all_instances = [self]
    all_instances.extend(composed_instances)
    for instance in all_instances:
        if instance.additional_properties_type is not None:
            additional_properties_model_instances.append(instance)
    return additional_properties_model_instances


def get_var_name_to_model_instances(self, composed_instances):
    var_name_to_model_instances = {}
    all_instances = [self]
    all_instances.extend(composed_instances)
    for instance in all_instances:
        for var_name in instance.openapi_types():
            if var_name not in var_name_to_model_instances:
                var_name_to_model_instances[var_name] = [instance]
            else:
                var_name_to_model_instances[var_name].append(instance)
    return var_name_to_model_instances


def get_unused_args(self, composed_instances, model_args):
    unused_args = dict(model_args)
    # arguments apssed to self were already converted to python names
    # before __init__ was called
    for var_name_py in self.attribute_map:
        if var_name_py in unused_args:
            del unused_args[var_name_py]
    for instance in composed_instances:
        if instance.__class__ in self._composed_schemas()['allOf']:
            for var_name_py in instance.attribute_map:
                if var_name_py in unused_args:
                    del unused_args[var_name_py]
        else:
            for var_name_js in instance.attribute_map.values():
                if var_name_js in unused_args:
                    del unused_args[var_name_js]
    return unused_args


def validate_get_composed_info(constant_args, model_args, self):
    """
    For composed schemas/classes, validates the classes to make sure that
    they do not share any of the same parameters. If there is no collision
    then composed model instances are created and returned tot the calling
    self model

    Args:
        constant_args (dict): these are the args that every model requires
        model_args (dict): these are the required and optional spec args that
            were passed in to make this model
        self (class): the class that we are instantiating
            This class contains self._composed_schemas()

    Returns:
        composed_info (list): length three
            composed_instances (list): the composed instances which are not
                self
            var_name_to_model_instances (dict): a dict going from var_name
                to the model_instance which holds that var_name
                the model_instance may be self or an instance of one of the
                classes in self.composed_instances()
            additional_properties_model_instances (list): a list of the
                model instances which have the property
                additional_properties_type. This list can include self
    """
    # create composed_instances
    composed_instances = []
    allof_instances = get_allof_instances(self, model_args, constant_args)
    composed_instances.extend(allof_instances)
    oneof_instance = get_oneof_instance(self, model_args, constant_args)
    if oneof_instance is not None:
        composed_instances.append(oneof_instance)
    anyof_instances = get_anyof_instances(self, model_args, constant_args)
    composed_instances.extend(anyof_instances)

    # map variable names to composed_instances
    var_name_to_model_instances = get_var_name_to_model_instances(
        self, composed_instances)

    # set additional_properties_model_instances
    additional_properties_model_instances = (
        get_additional_properties_model_instances(composed_instances, self)
    )

    # set any remaining values
    unused_args = get_unused_args(self, composed_instances, model_args)
    if len(unused_args) > 0 and \
            len(additional_properties_model_instances) == 0 and \
            (self._configuration is None or
                not self._configuration.discard_unknown_keys):
        raise ApiValueError(
            "Invalid input arguments input when making an instance of "
            "class %s. Not all inputs were used. The unused input data "
            "is %s" % (self.__class__.__name__, unused_args)
        )

    # no need to add additional_properties to var_name_to_model_instances here
    # because additional_properties_model_instances will direct us to that
    # instance when we use getattr or setattr
    # and we update var_name_to_model_instances in setattr

    return [
      composed_instances,
      var_name_to_model_instances,
      additional_properties_model_instances,
      unused_args
    ]



# coding: utf-8

"""
    LAMP Platform

    The LAMP Platform API.

    The version of the OpenAPI document: 1.0.0
    Contact: team@digitalpsych.org
"""

from setuptools import setup, find_packages

__version__ = "2021.1.13"

setup(
    name="LAMP_core",
    version=__version__,
    description="LAMP Platform",
    author="Division of Digital Psychiatry at Beth Israel Deaconess Medical Center.",
    author_email="team@digitalpsych.org",
    url="https://docs.lamp.digital/",
    keywords=["LAMP Platform"],
    install_requires=[
      "urllib3 >= 1.15",
      "six >= 1.10",
      "certifi",
      "python-dateutil",
      "nulltype",
      "numpy",
      "pandas",
    ],
    extras_require={':python_version <= "2.7"': ['future']},
    packages=find_packages(exclude=["test", "tests"]),
    include_package_data=True,
    long_description="""\
    The LAMP Platform API.
    """
)



import LAMP
import numpy as np
from functools import reduce

class StudyExt():
    """
    """
    def __init__(self, 
                 participants, 
                 domains=None, 
                 df_props={}):

        self.domains = domains
        self.df_props = df_props
        self.init_participants(participants)

    def __iter__(self):
        for participant in self.participants:
            yield participant

    def __len__(self):
        return len(self.participants)

    def __getitem__(self, key):
        return self.participants[key]

    @property
    def participants(self):
        return self._participants

    @property
    def domains(self):
        return self._domains

    @participants.setter
    def participants(self, value):
        self._participants = value

    @domains.setter
    def domains(self, value):
        self._domains = value

    def domain_check(self, domains):
        """
        If domains is passed in, just return it.

        Else, see if domains is set as object attribute
        """
        if domains is None:
            if not hasattr(self, 'domains'):
                raise AttributeError('Domains were not set for cohort and were not provided.')
            domains = self.domains
        return domains

    def init_participants(self, participants):
        """
        Initialize participants in cohorts. Can take either participant objects or participant ids
        """
        self.participants = []
        for participant in participants:

            if isinstance(participant, LAMP.analysis.ParticipantExt):
                #NEED TO CHECK participant MATCHES PROPS
                self.participants.append(participant)

            else:
                self.add_participant(participant)

    def get_participant(self, participant_id):
        """
        Get participant object by id.

        If it doesn't exist, will return None
        """
        for participant in self.participants:
            if participant.id == participant_id:
                return participant

        print('participant not found!')
        return None

    def add_participant(self, participant):
        self.participants.append(LAMP.analysis.ParticipantExt(id = participant, 
                                          domains=self.domains, 
                                          df_props=self.df_props))


    def mean_age(self):
        """
        Return mean, std age of participants in cohort
        """
        participant_ages = [participant.age for participant in self.participants if participant.age is not None]
        if len(participant_ages) == 0:
            return None

        return np.mean(participant_ages)


    def std_age(self):
        """
        Return std age of participants in cohort
        """
        participant_ages = [participant.age for participant in self.participants if participant.age is not None]
        if len(participant_ages) == 0:
            return None

        return np.std(participant_ages)

    def domain_mean(self, domain):
        """
        Find the mean value for particular domain in cohort
        
        :param domain (str): the specified domain
        """
        dom_values = [participant.df[domain].values for participant in self.participants if domain in participant.df.columns]
        if len(dom_values) == 0: return None
        return np.nanmean(np.concatenate(dom_values))

    def domain_stdev(self, domain):
        """
        Find the std value for particular domain in cohort
        
        :param domain (str): the specified domain
        """
        dom_values = [participant.df[domain].values for participant in self.participants if domain in participant.df.columns]
        if len(dom_values) == 0: return None
        return np.nanstd(np.concatenate(dom_values))

    def normalize(self, domains=None, dom_means={}, dom_vars={}, in_sample=False):
        """
        Normalize each domain in cohort so that values have 0 mean/unit variance

        If in_sample is true, then performs within-sample normalization
        
        :param domains (list): domains to use. Default None, in which all availble domains are used
        :param dom_means (dict): the predetermined means of specified domains
        :param dom_vars (dict): the predetermined standard deviations of specified domains
        :param in_sample (bool): Whether to perform within-sample normalization. Default False
        """
        if domains is None and self.domains is None: #Get all features from all particpants
            domains = set(np.concatenate([participant.df.columns.drop(['Date', 'id']) for participant in participants]))

        elif domains is None:
            domains = self.domains
            
        #
        if not in_sample:
            dom_means = {dom: self.domain_mean(dom) for dom in domains if self.domain_mean(dom) is not None and dom not in dom_means}
            dom_vars = {dom: self.domain_stdev(dom) for dom in domains if self.domain_stdev(dom) is not None and dom not in dom_vars}

        for participant in self.participants: participant.normalize(domains=domains, domain_means=dom_means, domain_vars=dom_vars)

    def impute(self, domains=None):
        """
        Impute every participant in cohort.
        
        :param domains (list): domains to use. Default None, in which all availble domains are used
        """
        if domains is None and self.domains is None: #Get all features from all particpants
            domains = set(np.concatenate([participant.df.columns.drop(['Date', 'id']) for participant in participants]))
        elif domains is None:
            domains = self.domains

        for participant in self: participant.impute(domains=domains)

    def bin(self, domains=None, window_size=3):
        """
        Bins all participants in cohort.
        
        :param domains (list): domains to use. Default None, in which all availble domains are used
        :param window_size (int): the number of days to use per bin
        """
        if domains is None and self.domains is None: #Get all features from all particpants
            domains = set(np.concatenate([participant.df.columns.drop(['Date', 'id']) for participant in participants]))
        elif domains is None:
            domains = self.domains

        for participant in self: participant.bin(domains=domains, window_size=window_size)
            
    def impute_bins(self, domains=None):
        """
        Impute bins
        
        :param domains (list): domains to use. Default None, in which all availble domains are used
        """
        if domains is None and self.domains is None: #Get all features from all particpants
            domains = set(np.concatenate([participant.df.columns.drop(['Date', 'id']) for participant in participants]))
        elif domains is None:
            domains = self.domains
            
        for participant in self: participant.impute_bins(domains=domains)
        

    def transition_probabilities(self, domains=None, joint_size=1):
        """
        Get cohort_wide transistion probabilities.

        :param domains (list): domains to use. Default None, in which all availble domains are used
        :param joint_size (int): the number of variables used when calculating the joint probabilities for transistion event. Defaults to 1. 
        """
        if domains is None and self.domains is None: #Get all features from all particpants
            domains = set(np.concatenate([participant.df.columns.drop(['Date', 'id']) for participant in participants]))
        elif domains is None:
            domains = self.domains

        samples_tp = [pro.get_transitions(domains = domains, joint_size = joint_size) for pro in self]

        master_dict = {}

        for sample in samples_tp:
            for cat in sample:
                if cat not in master_dict:
                    master_dict[cat] = {state: sample[cat][state] for state in sample[cat]}
                else: #merge
                    for state in sample[cat]:
                        master_dict[cat][state] = {state2: master_dict[cat][state][state2] + sample[cat][state][state2] for state2 in 
                                                                                                                        master_dict[cat][state]}
        #Convert to probabilities
        trans_dict = {}
        for cat in master_dict:
            trans_dict[cat] = {}
            for state in master_dict[cat]:
                trans_dict[cat][state] = {}
                for state2 in master_dict[cat][state]:
                    if sum(master_dict[cat][state].values()) == 0:
                        trans_dict[cat][state][state2] = None
                    else:
                        trans_dict[cat][state][state2] = float(master_dict[cat][state][state2]) / float(sum(master_dict[cat][state].values()))
        return trans_dict, master_dict

    def domain_bouts(self, domains=None):
        """
        Get elevated/sedated domain bouts in each domain
        
        :param domains (list): domains to use. Default None, in which all availble domains are used
        """
        if domains is None and self.domains is None: #Get all features from all particpants
            domains = set(np.concatenate([participant.df.columns.drop(['Date', 'id']) for participant in participants]))
        elif domains is None:
            domains = self.domains
            
        bout_dict = {}
        for participant in self: 
            participant_bout_dict = participant.domain_bouts(domains=domains)
            for dom in participant_bout_dict:
                if dom not in bout_dict: bout_dict[dom] = participant_bout_dict[dom]
                else: 
                    bout_dict[dom]['low'] += participant_bout_dict[dom]['low']
                    bout_dict[dom]['high'] += participant_bout_dict[dom]['high']
                    bout_dict[dom]['low ends'] += participant_bout_dict[dom]['low ends']
                    bout_dict[dom]['high ends'] += participant_bout_dict[dom]['high ends']

        return bout_dict



